#include <iostream>

#include <cstdlib>
#include <cstring>
#include <ctime>

using namespace std;

#include "common.h"
#include "logdbfile.h"
#include "logdb.h"

void LogDB::WriteIndex(LONGINT start_ptr, LONGINT num_ptr)
{
    // write ptr blocks
    int noblocks = (num_ptr - 1) / ptr_per_block + 1;
    LONGINT* ptrs = new LONGINT[noblocks * ptr_per_block];
    INT ix = 0;
    for(LONGINT ptr = start_ptr; ptr < start_ptr + num_ptr; ptr++)
    {
        ptrs[ix] = ptr;
        ix++;
    }
    while(ix < noblocks * ptr_per_block)
    {
        ptrs[ix] = UNUSED;
        ix++;
    }
    dbf->Write(ptrs, noblocks);
    delete[] ptrs;
    // if more than 1 ptr block write another level of ptr blocks
    if(noblocks > 1)
    {
        WriteIndex(start_ptr + num_ptr, noblocks);
    }
}

void LogDB::DumpIndex(LONGINT ptr, DUMP_DATA_F dump_data, INT levels_left, ostream& os)
{
    os << "Index Page " << ptr << ":" << endl;
    LONGINT* ptrs = new LONGINT[ptr_per_block];
    dbf->Read(ptr, ptrs);
    for(INT ix = 0; ix < ptr_per_block; ix++)
    {
        if(ptrs[ix] != UNUSED)
        {
            os << "  Page: " << ptrs[ix] << endl;
        }
    }
    for(INT ix = 0; ix < ptr_per_block; ix++)
    {
        if(ptrs[ix] >= 0)
        {
            if(levels_left > 1)
            {
                DumpIndex(ptrs[ix], dump_data, levels_left - 1, os);
            }
            else
            {
                DumpData(ptrs[ix], dump_data, os);
            }
        }
    }
    delete[] ptrs;
}

void LogDB::DumpData(LONGINT ptr, DUMP_DATA_F dump_data, ostream& os)
{
    os << "Data Page " << ptr << ":" << endl;
    char* buf = new char[block_size];
    dbf->Read(ptr, buf);
    for(int ix = 0; ix < data_per_block; ix++)
    {
        os << "  " << dump_data(buf + ix * data_size) << endl;
    }
    delete [] buf;
}

void LogDB::Load(LONGINT elm, time_t t, LONGINT* ptrs, void* buf, int* ix_offset, int *data_offset)
{
    LONGINT ptr = dbf->GetRoot(t);
    LONGINT tmp = elm;
    for(int lvl = levels - 1; lvl >= 0; lvl--)
    {
        dbf->Read(ptr, ptrs + lvl * ptr_per_block);
        ptr = ptrs[lvl * ptr_per_block + tmp / steps[lvl]];
        ix_offset[lvl] = tmp / steps[lvl];
        tmp = tmp % steps[lvl];
    }
    dbf->Read(ptr, buf);
    *data_offset = tmp;
}

void LogDB::BatchLoad(LONGINT elm, time_t t, LONGINT* ptrs, void* buf, int* ix_offset, int *data_offset, LONGINT* old_ix_ptr, LONGINT *old_data_ptr)
{
    LONGINT ptr = dbf->GetRoot(t);
    LONGINT tmp = elm;
    for(int lvl = levels - 1; lvl >= 0; lvl--)
    {
        old_ix_ptr[lvl] = ptr;
        if(!cache->Get(lvl, ptr,  ptrs + lvl * ptr_per_block))
        {
            dbf->Read(ptr, ptrs + lvl * ptr_per_block);
        }
        ptr = ptrs[lvl * ptr_per_block + tmp / steps[lvl]];
        ptr = abs(ptr); // handle negative "updated" values
        ix_offset[lvl] = tmp / steps[lvl];
        tmp = tmp % steps[lvl];
    }
    *old_data_ptr = ptr;
    if(!cache->Get(-1, ptr, buf))
    {
        dbf->Read(ptr, buf);
    }
    *data_offset = tmp;
}

LogDB::LogDB(LONGINT total_data, INT data_size, INT data_per_block)
{
    this->data_size = data_size;
    this->data_per_block = data_per_block;
    this->total_data = total_data;
    this->block_size = data_per_block * data_size;
    if(this->block_size % sizeof(LONGINT) != 0)
    {
        throw "Block size not multipla of pointer size";
    }
    this->ptr_per_block = this->block_size / sizeof(LONGINT);
    this->total_data_blocks = (total_data - 1) / data_per_block + 1;
    this->levels = 0;
    INT leafs = this->total_data_blocks;
    while(leafs > 1)
    {
        this->levels++;
        leafs = (leafs - 1) / this->ptr_per_block + 1;
    }
    steps = new INT[this->levels];
    steps[0] = data_per_block;
    for(int i = 1; i < levels; i++)
    {
        steps[i] = ptr_per_block * steps[i-1];
    }
}

LogDB::~LogDB()
{
    delete dbf;
    delete[] steps;
}

void LogDB::Init(const char* db_file_name, const char* ix_file_name, GET_DATA_F get_data, int cache_entries)
{
    // create new file
    dbf = new LogDBFile(db_file_name, ix_file_name, "wb+", block_size, cache_entries);
    // write data blocks
    char* buf = new char[block_size];
    for(LONGINT dptr = 0; dptr < total_data_blocks; dptr++)
    {
        memset(buf, 0, block_size);
        for(INT ix = 0; ix < data_per_block; ix++)
        {
            if(dptr * data_per_block + ix < total_data)
            {
                get_data(dptr * data_per_block + ix, buf + ix * data_size);
            }
        }
        dbf->Write(buf);
    }
    delete[] buf;
    // write index blocks
    WriteIndex(0, total_data_blocks);
    // start timestamp
    dbf->Mark();
}

void LogDB::Open(const char* db_file_name, const char* ix_file_name, int cache_entries)
{
    // open existing file
    dbf = new LogDBFile(db_file_name, ix_file_name, "rb+", block_size, cache_entries);
}

LONGINT LogDB::Size()
{
    return (dbf->GetLastWritePtr() + 1) * block_size;
}

void LogDB::Dump(DUMP_DATA_F dump_data, ostream& os)
{
    Dump(dump_data, NOW, os);
}

void LogDB::Dump(DUMP_DATA_F dump_data, time_t t, ostream& os)
{
    DumpIndex(dbf->GetRoot(t), dump_data, levels, os);
}

#define LOADINIT LONGINT* ptrs = new LONGINT[levels * ptr_per_block]; \
                 char* buf = new char[block_size]; \
                 int* ix_offset = new int[levels]; \
                 int data_offset;

#define LOADEND delete[] ptrs; \
                delete [] buf; \
                delete[] ix_offset;

void LogDB::Read(LONGINT elm, void* res)
{
    Read(elm, res, NOW);
}

void LogDB::Read(LONGINT elm, void* res, time_t t)
{
    LOADINIT
    Load(elm, t, ptrs, buf, ix_offset, &data_offset);
    memcpy(res, buf + data_offset * data_size, data_size);
    LOADEND
}

void LogDB::Update(LONGINT elm, void* res)
{
    LOADINIT
    Load(elm, time(NULL), ptrs, buf, ix_offset, &data_offset);
    // update data
    memcpy(buf + data_offset * data_size, res, data_size);
    dbf->Write(buf);
    // update parent indexes
    for(int lvl = 0; lvl < levels; lvl++)
    {
        ptrs[lvl * ptr_per_block + ix_offset[lvl]] = dbf->GetLastWritePtr();
        dbf->Write(&ptrs[lvl * ptr_per_block]);
    }
    // save new root
    dbf->Mark();
    LOADEND
}

void LogDB::BatchUpdateBegin()
{
    cache = new WriteCache(block_size);
}

void LogDB::BatchUpdate(LONGINT elm, void* res)
{
    LOADINIT
    LONGINT* old_ix_ptr = new LONGINT[levels];
    LONGINT old_data_ptr;
    BatchLoad(elm, time(NULL), ptrs, buf, ix_offset, &data_offset, old_ix_ptr, &old_data_ptr);
    // update data
    memcpy(buf + data_offset * data_size, res, data_size);
    cache->Insert(-1, old_data_ptr, buf);
    // update parent indexes
    for(int lvl = 0; lvl < levels; lvl++)
    {
        ptrs[lvl * ptr_per_block + ix_offset[lvl]] = -(lvl > 0 ? old_ix_ptr[lvl - 1] : old_data_ptr); // mark pointers to be updated later by negating them
        cache->Insert(lvl, old_ix_ptr[lvl], &ptrs[lvl * ptr_per_block]);
    }
    LOADEND
    delete[] old_ix_ptr;
}

void LogDB::BatchUpdateCommit()
{
    char* buf = new char[block_size];
    map<LONGINT, LONGINT> old2new;
    for(int lvl = -1; lvl < levels; lvl++)
    {
        vector<LONGINT> ptrs;
        cache->GetAll(lvl, ptrs);
        vector<LONGINT>::iterator ptr = ptrs.begin();
        while(ptr != ptrs.end())
        {
            // get cache data
            cache->Get(lvl, *ptr, buf);
            // if index then update negative pointers
            if(lvl > -1)
            {
                LONGINT* bufptrs = (LONGINT*)buf;
                for(int i = 0; i < ptr_per_block; i++)
                {
                    if(bufptrs[i] < 0 && bufptrs[i] != UNUSED)
                    {
                        bufptrs[i] = old2new[-bufptrs[i]];
                    }
                }
            }
            // write to disk
            dbf->Write(buf);
            // update mapping from old to new pointers
            old2new[*ptr] = dbf->GetLastWritePtr();
            ptr++;
        }
    }
    delete[] buf;
    dbf->Mark();
    delete cache;
}
