#include <iostream>
#include <iomanip>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>

using namespace std;

#include "logdb.h"

const char* LOGDB = "test2.logdb";
const char* IX = "test2.ix";

struct Rec
{
    double xv;
};

void get(INT ix, void* buf)
{
    Rec r;
    r.xv = ix;
    memcpy(buf, &r, sizeof(Rec));
}

char res[100];

char* dump(void* buf)
{
    Rec* r = (Rec*)buf;
    sprintf(res, "%f", r->xv);
    return res;
}

const int NR = 1 << 24;
const int NW = 1 << 16; // must be power of 2 for LCG to work

void Check(LogDB& db, int delta, bool wrt)
{
    for(int i = 0; i < (wrt ? NW : NR); i++)
    {
        Rec r;
        db.Read(i, &r);
        if((int)r.xv != i + delta)
        {
            cout << "Problem for element " << i << ": " << r.xv << endl;
        }
    }
}

int LCG(int ix)
{
    return (5 * ix + 1) % NW;
}

void Create(bool wrt)
{
    LogDB db(wrt ? NW : NR, sizeof(Rec), 10);
    db.Init(LOGDB, IX, get);
}

void TestRead(int cache)
{
    LogDB db(NR, sizeof(Rec), 10);
    db.Open(LOGDB, IX, cache);
    time_t t1 = time(NULL);
    Check(db, 0, false);
    time_t t2 = time(NULL);
    if(t2 == t1) t2++;
    cout << "read speed: " << (NR / (t2 - t1)) << " rec/sec with cache " << cache << endl;
}

void TestWrite()
{
    LogDB db(NW, sizeof(Rec), 10);
    db.Open(LOGDB, IX, 100);
    LONGINT s1 = db.Size();
    time_t t1 = time(NULL);
    int ix = 0;
    for(int i = 0; i < NW; i++)
    {
        Rec r;
        db.Read(ix, &r);
        r.xv = r.xv + 1;
        db.Update(ix, &r);
        ix = LCG(ix);
    }
    time_t t2 = time(NULL);
    if(t2 == t1) t2++;
    LONGINT s2 = db.Size();
    cout << "write speed: " << (NW / (t2 - t1)) << " rec/sec" << endl;
    cout << "size overhead: " << fixed << setprecision(1) << ((s2 - s1) * 1.0 / (NW * sizeof(Rec))) << endl;
    Check(db, 1, true);
}

void TestBatchWrite(int batch, int delta)
{
    if(NW % batch != 0) throw "Batch size not multipla of number of entries";
    cout << "batch size: " << batch << endl;
    LogDB db(NW, sizeof(Rec), 10);
    db.Open(LOGDB, IX, 100);
    LONGINT s1 = db.Size();
    time_t t1 = time(NULL);
    int ix = 0;
    db.BatchUpdateBegin();
    for(int i = 0; i < NW; i++)
    {
        Rec r;
        db.Read(ix, &r);
        r.xv = r.xv + 1;
        db.BatchUpdate(ix, &r);
        ix = LCG(ix);
        if((i + 1) % batch == 0)
        {
            db.BatchUpdateCommit();
            db.BatchUpdateBegin();
        }
    }
    db.BatchUpdateCommit();
    time_t t2 = time(NULL);
    if(t2 == t1) t2++;
    LONGINT s2 = db.Size();
    cout << "write speed: " << (NW / (t2 - t1)) << " rec/sec" << endl;
    cout << "size overhead: " << fixed << setprecision(1) << ((s2 - s1) * 1.0 / (NW * sizeof(Rec))) << endl;
    Check(db, delta, true);
}

const int REP = 3;

int main()
{
    try
    {
        Create(false);
        for(int i = 0; i < REP; i++)
        {
            TestRead(0);
            TestRead(10);
            TestRead(100);
        }
        Create(true);
        TestWrite();
        TestBatchWrite(1 << 4, 2);
        TestBatchWrite(1 << 8, 3);
        TestBatchWrite(1 << 12, 4);
    }
    catch(char const* ex)
    {
        cerr << ex << endl;
    }
    return EXIT_SUCCESS;
}
